/*
 *  linux/drivers/fusion/mpilan.c
 *      IP Over Fibre Channel device driver.
 *      For use with PCI adapter(s):
 *          LSI SYM409xx (Fibre Channel)
 *      running LSI Logic Fusion MPT firmware.
 *
 *  Credits:
 *      This driver would not exist if not for Alan Cox's development
 *      of the linux i2o driver.
 *
 *      Special thanks goes to the I2O LAN driver people at the
 *      University of Helsinki, who, unbeknownst to them, provided
 *      the inspiration and initial structure for this driver.
 *
 *      A huge debt of gratitude is owed to David S. Miller (DaveM)
 *      for fixing much of the stupid and broken stuff in the early
 *      driver while porting to sparc64 platform.  THANK YOU!
 *
 *      A really huge debt of gratitude is owed to Eddie C. Dost
 *      for gobs of hard work fixing and optimizing LAN code.
 *      THANK YOU!
 *
 *  Copyright (c) 2000 LSI Logic Corporation
 *  Originally By: Noah Romer
 *
 *  $Id: mpilan.c,v 1.55 2000/09/29 23:35:53 nromer Exp $
 */
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    NO WARRANTY
    THE PROGRAM IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OR
    CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED INCLUDING, WITHOUT
    LIMITATION, ANY WARRANTIES OR CONDITIONS OF TITLE, NON-INFRINGEMENT,
    MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. Each Recipient is
    solely responsible for determining the appropriateness of using and
    distributing the Program and assumes all risks associated with its
    exercise of rights under this Agreement, including but not limited to
    the risks and costs of program errors, damage to or loss of data,
    programs or equipment, and unavailability or interruption of operations.

    DISCLAIMER OF LIABILITY
    NEITHER RECIPIENT NOR ANY CONTRIBUTORS SHALL HAVE ANY LIABILITY FOR ANY
    DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING WITHOUT LIMITATION LOST PROFITS), HOWEVER CAUSED AND
    ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
    USE OR DISTRIBUTION OF THE PROGRAM OR THE EXERCISE OF ANY RIGHTS GRANTED
    HEREUNDER, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 * Define statements used for debuging
 */
//#define MPI_LAN_IO_DEBUG

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#include "mpilan.h"
#include <linux/fs.h>

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 * MPT LAN message sizes without variable part.
 */
#define MPI_LAN_RECEIVE_POST_REQUEST_SIZE \
	(sizeof(LANReceivePostRequest_t) - sizeof(SGE_MPI_UNION))

#define MPI_LAN_TRANSACTION32_SIZE \
	(sizeof(SGETransaction32_t) - sizeof(U32))

/*
 *  Fusion MPT LAN private structures
 */

struct BufferControl {
	struct sk_buff	*skb;
	dma_addr_t	dma;
	unsigned int	len;
};

struct mpi_lan_priv {
	MPI_ADAPTER *mpi_dev;
	u8 pnum; /* Port number in the IOC. This is not a Unix network port! */

	atomic_t buckets_out;	  	/* number of unused buckets on IOC */
	int bucketthresh;		/* Send more when this many used */

	int *mpi_txfidx; /* Free Tx Context list */
	int mpi_txfidx_tail;
	spinlock_t txfidx_lock;

	int *mpi_rxfidx; /* Free Rx Context list */
	int mpi_rxfidx_tail;
	spinlock_t rxfidx_lock;

	struct BufferControl *RcvCtl;	/* Receive BufferControl structs */
	struct BufferControl *SendCtl;	/* Send BufferControl structs */

	int max_buckets_out;	   	/* Max buckets to send to IOC */
	int tx_max_out;		   	/* IOC's Tx queue len */

	u32 total_posted;
	u32 total_received;
	struct net_device_stats stats;	/* Per device statistics */

	struct tq_struct post_buckets_task;
	unsigned long post_buckets_active;
};

struct mpi_lan_ohdr {
	u16	dtype;
	u8	daddr[FC_ALEN];
	u16	stype;
	u8	saddr[FC_ALEN];
};

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*
 *  Forward protos...
 */
STATIC int  lan_reply (MPI_ADAPTER *ioc, MPI_FRAME_HDR *mf,
		       MPI_FRAME_HDR *reply);
STATIC int  mpi_lan_open(struct net_device *dev);
STATIC int  mpi_lan_reset(struct net_device *dev);
STATIC int  mpi_lan_close(struct net_device *dev);
STATIC void mpi_lan_post_receive_buckets(void *dev_id);
STATIC void mpi_lan_wake_post_buckets_task(struct net_device *dev);
STATIC int  mpi_lan_receive_post_turbo(struct net_device *dev, u32 tmsg);
STATIC int  mpi_lan_receive_post_reply(struct net_device *dev,
				       LANReceivePostReply_t *pRecvRep);
STATIC int  mpi_lan_send_turbo(struct net_device *dev, u32 tmsg);
STATIC int  mpi_lan_send_reply(struct net_device *dev,
			       LANSendReply_t *pSendRep);
STATIC int  mpi_lan_event_notification(struct net_device *dev, U8 Switch);
STATIC void mpi_lan_event_reply(struct net_device *dev,
				EventNotificationReply_t *pEvent);
STATIC unsigned short mpi_lan_type_trans(struct sk_buff *skb,
					 struct net_device *dev);

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Fusion MPT LAN private data
 */
static int LanCtx = -1;

static u32 max_buckets_out = 127;
static u32 tx_max_out_p = 127 - 16;

static struct net_device *mpi_landev[myMPI_MAX_ADAPTERS+1];

#define mpi_lan_init init_module

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 * Fusion MPT LAN external data
 */
extern int LanCIdx;

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 * lan_reply(): Handle all data sent from the hardware.
 */
STATIC int
lan_reply (MPI_ADAPTER *ioc, MPI_FRAME_HDR *mf, MPI_FRAME_HDR *reply)
{
	struct net_device *dev = mpi_landev[ioc->id];
	int FreeReqFrame = 0;

	dioprintk ((KERN_INFO "mpilan: Got reply. ioc %d, dev %s.\n",
		  ioc->id, dev->name));

//	dioprintk ((KERN_INFO "mpilan/lan_reply: mf %p, reply %p.\n", mf, reply));

	if (mf == NULL) {
#if defined(__alpha__) || defined(__sparc_v9__)
		u32 tmsg = (u32) ((u64)reply);
#else
		u32 tmsg = (u32) reply;
#endif

		dioprintk((KERN_INFO "lan_reply: tmsg %08x.\n", tmsg));

		switch (tmsg & 0xF8000000) {
		case 0x58000000:
//			dioprintk ((KERN_INFO "mpilan/lan_reply: "
//				  "MessageContext turbo reply received.\n"));
			FreeReqFrame = 1;
			break;

		case 0x50000000:
//			dioprintk (("mpilan/lan_reply: "
//				  "calling mpi_lan_send_reply (turbo).\n"));
			FreeReqFrame = mpi_lan_send_turbo(dev, tmsg);
			break;

		case 0x40000000:
//			dioprintk ((KERN_INFO "mpilan/lan_reply: "
//				  "rcv-Turbo = %08x.\n", tmsg));
			mpi_lan_receive_post_turbo(dev, tmsg);
			break;

		default:
			printk (KERN_ERR "mpilan/lan_reply: Got a turbo reply "
				"that I don't know what to do with.\n");
			break;
		}

		return FreeReqFrame;
	}

//	msg = (u32 *) reply;
//	dioprintk ((KERN_INFO "mpilan/lan_reply: "
//		  "msg: %08x %08x %08x %08x\n",
//		  le32_to_cpu(msg[0]), le32_to_cpu(msg[1]),
//		  le32_to_cpu(msg[2]), le32_to_cpu(msg[3])));
//	dioprintk ((KERN_INFO "lan_reply: Function 0x%02x.\n",
//		  reply->u.hdr.Function));

	switch (reply->u.hdr.Function) {

	case MPI_FUNCTION_LAN_SEND:
	{
		LANSendReply_t *pSendRep;

		pSendRep = (LANSendReply_t *) reply;
		FreeReqFrame = mpi_lan_send_reply(dev, pSendRep);
		break;
	}

	case MPI_FUNCTION_LAN_RECEIVE:
	{
		LANReceivePostReply_t *pRecvRep;

		pRecvRep = (LANReceivePostReply_t *) reply;
		if (pRecvRep->NumberOfContexts) {
			mpi_lan_receive_post_reply(dev, pRecvRep);
			if (!(pRecvRep->MsgFlags & MPI_MSGFLAGS_CONTINUATION_REPLY))
				FreeReqFrame = 1;
		} else
			dioprintk ((KERN_INFO "lan_reply: zero context "
				  "ReceivePostReply received.\n"));
		break;
	}

	case MPI_FUNCTION_LAN_RESET:
		/* Just a default reply. Might want to check it to
		 * make sure that everything went ok.
		 */
		FreeReqFrame = 1;
		break;

	case MPI_FUNCTION_EVENT_NOTIFICATION:
	{
		EventNotificationReply_t *pEvent;

		pEvent = (EventNotificationReply_t *) reply;
		mpi_lan_event_reply(dev, pEvent);
		break;
	}

	default:
		printk (KERN_ERR "mpilan/lan_reply: Got a non-turbo "
			"reply that I don't know what to do with.\n");
		break;
	}

	return FreeReqFrame;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC void
mpi_lan_event_reply(struct net_device *dev, EventNotificationReply_t *pEvent)
{
	MPI_FRAME_HDR *mf;
	EventAck_t *pAck;
	U32 data;
	int i = 0;

	printk (KERN_INFO "%s: MPI event - ", dev->name);
	
	switch (le32_to_cpu(pEvent->Event)) {

	case MPI_EVENT_NONE:
		printk ("None:");
		break;

	case MPI_EVENT_LOG_DATA:
		printk ("Log Data:");
		break;

	case MPI_EVENT_STATE_CHANGE:
		printk ("State Change:");
		break;

	case MPI_EVENT_UNIT_ATTENTION:
		printk ("Unit Attention:");
		break;

	case MPI_EVENT_IOC_BUS_RESET:
		printk ("IOC Bus Reset:");
		break;

	case MPI_EVENT_EXT_BUS_RESET:
		printk ("External Bus Reset:");
		break;

	case MPI_EVENT_RESCAN:
		printk ("Bus Rescan Event:"); 
		/* Ok, do we need to do anything here? As far as
		   I can tell, this is when a new device gets added
		   to the loop. */
		break;

	case MPI_EVENT_LINK_STATUS_CHANGE:
		printk ("Link Status Change:");

		data = le32_to_cpu(pEvent->Data[0]);
		if (data == MPI_EVENT_LINK_STATUS_FAILURE)
			printk (" Failure:");
		else
			printk (" Active:");
		break;

	case MPI_EVENT_LOOP_STATE_CHANGE:
		printk ("Loop State Change:");

		data = le32_to_cpu(pEvent->Data[0]);
		if (data == MPI_EVENT_LOOP_STATE_CHANGE_LIP)
			printk (" LIP:");
		else if (data == MPI_EVENT_LOOP_STATE_CHANGE_LPE)
			printk (" LPE:"); /* ??? */
		else
			printk (" LPB:"); /* ??? */
		break;

	case MPI_EVENT_LOGOUT:
		printk ("Logout:");
		break;

	case MPI_EVENT_EVENT_CHANGE:
		printk ("Event Change:");
		break;

	default:
		printk ("Unknown %08x:", le32_to_cpu(pEvent->Event));
		break;
	}

	for (i = 0; i < le16_to_cpu(pEvent->EventDataLength); i++)
		printk (" %08x", le32_to_cpu(pEvent->Data[i]));
	printk ("\n");

	if (pEvent->AckRequired == MPI_EVENT_NOTIFICATION_ACK_REQUIRED) {
		struct mpi_lan_priv *priv = dev->priv;
		int id = priv->mpi_dev->id;

		mf = MpiGetMsgFrame (LanCtx, id);
		if (mf == NULL) {
/*			dprintk ((KERN_ERR
				  "mpilan/event_reply: Evil funkiness abounds! "
				  "Unable to allocate a request frame.\n"));
*/
			return;
		}
		pAck = (EventAck_t *) mf;

		pAck->Function     = MPI_FUNCTION_EVENT_ACK;
		pAck->ChainOffset  = 0;
		pAck->MsgFlags     = 0;
		pAck->Event        = pEvent->Event;
		pAck->EventContext = pEvent->EventContext;

		MpiPutMsgFrame (LanCtx, id, mf);
	}
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC int
mpi_lan_event_notification(struct net_device *dev, U8 Switch)
{
	struct mpi_lan_priv *priv = (struct mpi_lan_priv *) dev->priv;
	MPI_ADAPTER *mpi_dev = priv->mpi_dev;
	EventNotification_t *pEvent;
	MPI_FRAME_HDR *mf;

	mf = MpiGetMsgFrame(LanCtx, mpi_dev->id);
	if (mf == NULL) {
		printk (KERN_ERR "%s: Unable to alloc request frame.\n",
			__FUNCTION__);
		return -1;
	}
	pEvent = (EventNotification_t *) mf;

	pEvent->Function    = MPI_FUNCTION_EVENT_NOTIFICATION;
	pEvent->ChainOffset = 0;
	pEvent->MsgFlags    = 0;
	pEvent->Switch      = Switch;

	MpiPutMsgFrame(LanCtx, mpi_dev->id, mf);
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC int
mpi_lan_open(struct net_device *dev)
{
	struct mpi_lan_priv *priv = (struct mpi_lan_priv *) dev->priv;
	int i;

	mpi_lan_reset(dev);

	priv->mpi_txfidx = kmalloc(priv->tx_max_out * sizeof(int), GFP_KERNEL);
	if (priv->mpi_txfidx == NULL)
		return -ENOMEM;
	priv->mpi_txfidx_tail = -1;

	priv->SendCtl = kmalloc(priv->tx_max_out * sizeof(struct BufferControl),
				GFP_KERNEL);
	if (priv->SendCtl == NULL)
		return -ENOMEM;
	for (i = 0; i < priv->tx_max_out; i++) {
		memset(&priv->SendCtl[i], 0, sizeof(struct BufferControl));
		priv->mpi_txfidx[++priv->mpi_txfidx_tail] = i;
	}

	dprintk ((KERN_INFO "mpilan/lo: Finished initializing SendCtl.\n"));

	priv->mpi_rxfidx = kmalloc(priv->max_buckets_out * sizeof(int),
				   GFP_KERNEL);
	if (priv->mpi_rxfidx == NULL)
		return -ENOMEM;
	priv->mpi_rxfidx_tail = -1;

	priv->RcvCtl = kmalloc(priv->max_buckets_out *
						sizeof(struct BufferControl),
			       GFP_KERNEL);
	if (priv->RcvCtl == NULL)
		return -ENOMEM;
	for (i = 0; i < priv->max_buckets_out; i++) {
		memset(&priv->RcvCtl[i], 0, sizeof(struct BufferControl));
		priv->mpi_rxfidx[++priv->mpi_rxfidx_tail] = i;
	}

/**/	dprintk ((KERN_INFO "mpilan/lo: txfidx contains - "));
/**/	for (i = 0; i < priv->tx_max_out; i++)
/**/		dprintk ((" %xh", priv->mpi_txfidx[i]));
/**/	dprintk (("\n"));

	dprintk ((KERN_INFO "mpilan/lo: Finished initializing RcvCtl.\n"));

	mpi_lan_post_receive_buckets(dev);

	if (mpi_lan_event_notification(dev, MPI_EVENT_NOTIFICATION_SWITCH_ON))
		printk (KERN_WARNING "%s: Unable to set the event mask.\n",
			dev->name);

	netif_start_queue(dev);

	MOD_INC_USE_COUNT;

	dprintk ((KERN_INFO "mpilan/lo: Done.\n"));

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC int
mpi_lan_reset(struct net_device *dev)
{
	MPI_FRAME_HDR *mf;
	LANResetRequest_t *pResetReq;
	struct mpi_lan_priv *priv = (struct mpi_lan_priv *)dev->priv;

	mf = MpiGetMsgFrame(LanCtx, priv->mpi_dev->id);

	if (mf == NULL) {
/*		dprintk ((KERN_ERR "mpilan/reset: Evil funkiness abounds! "
		"Unable to allocate a request frame.\n"));
*/
		return -1;
	}

	pResetReq = (LANResetRequest_t *) mf;

	pResetReq->Function	= MPI_FUNCTION_LAN_RESET;
	pResetReq->ChainOffset = 0;
	pResetReq->Reserved	= 0;
	pResetReq->PortNumber  = priv->pnum;
	pResetReq->MsgFlags	= 0;
	pResetReq->Reserved2   = 0;

	MpiPutMsgFrame(LanCtx, priv->mpi_dev->id, mf);
	
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC int
mpi_lan_close(struct net_device *dev)
{
	struct mpi_lan_priv *priv = (struct mpi_lan_priv *) dev->priv;
	MPI_ADAPTER *mpi_dev = priv->mpi_dev;
	unsigned int timeout;
	int i;

	if (mpi_lan_event_notification(dev, MPI_EVENT_NOTIFICATION_SWITCH_OFF))
		printk (KERN_WARNING "%s: Unable to clear the event mask.\n",
			dev->name);

	printk (KERN_INFO "mpilan: mpi_lan_close called.\n");
	dprintk ((KERN_INFO "mpilan:lan_close: Posted %d buckets "
		  "since driver was loaded, %d still out.\n",
		  priv->total_posted,atomic_read(&priv->buckets_out)));

	netif_stop_queue(dev);

	mpi_lan_reset(dev);

	timeout = 2 * HZ;
	while (atomic_read(&priv->buckets_out) && --timeout) {
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(1);
	}

	for (i = 0; i < priv->max_buckets_out; i++) {
		if (priv->RcvCtl[i].skb != NULL) {
/**/			dprintk ((KERN_INFO "mpilan/lan_close: bucket %05x "
/**/				  "is still out.\n", i));
			pci_unmap_single(mpi_dev->pcidev, priv->RcvCtl[i].dma,
				 	 priv->RcvCtl[i].len,
					 PCI_DMA_FROMDEVICE);
 			dev_kfree_skb(priv->RcvCtl[i].skb);
		}
	}

	kfree (priv->RcvCtl);
	kfree (priv->mpi_rxfidx);

	for (i = 0; i < priv->tx_max_out; i++) {
 		if (priv->SendCtl[i].skb != NULL) {
			pci_unmap_single(mpi_dev->pcidev, priv->SendCtl[i].dma,
					 priv->SendCtl[i].len,
					 PCI_DMA_TODEVICE);
 			dev_kfree_skb(priv->SendCtl[i].skb);
		}
	}

	kfree(priv->SendCtl);
	kfree(priv->mpi_txfidx);

	atomic_set(&priv->buckets_out, 0);

	MOD_DEC_USE_COUNT;
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC struct net_device_stats *
mpi_lan_get_stats(struct net_device *dev)
{
	struct mpi_lan_priv *priv = (struct mpi_lan_priv *)dev->priv;	

	return ((struct net_device_stats *) &priv->stats);
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC int
mpi_lan_change_mtu(struct net_device *dev, int new_mtu)
{
	if ((new_mtu < MPI_LAN_MIN_MTU) || (new_mtu > MPI_LAN_MAX_MTU))
		return -EINVAL;
	dev->mtu = new_mtu;
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/* Tx timeout handler. */
STATIC void
mpi_lan_tx_timeout(struct net_device *dev)
{
	netif_wake_queue(dev);  
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
//STATIC inline int
STATIC int
mpi_lan_send_turbo(struct net_device *dev, u32 tmsg)
{
	struct mpi_lan_priv *priv = (struct mpi_lan_priv *) dev->priv;
	MPI_ADAPTER *mpi_dev = priv->mpi_dev;
	struct sk_buff *sent;
	unsigned long flags;
	u32 ctx;

	ctx = tmsg & 0x07FFFFFF;
	sent = priv->SendCtl[ctx].skb;

	priv->stats.tx_packets++;
	priv->stats.tx_bytes += sent->len;

	dioprintk((KERN_INFO "%s: skb %p sent.\n", __FUNCTION__, sent));

	priv->SendCtl[ctx].skb = NULL;
	pci_unmap_single(mpi_dev->pcidev, priv->SendCtl[ctx].dma,
			 priv->SendCtl[ctx].len, PCI_DMA_TODEVICE);
	dev_kfree_skb_irq(sent);

	spin_lock_irqsave(&priv->txfidx_lock, flags);
	priv->mpi_txfidx[++priv->mpi_txfidx_tail] = ctx;
	spin_unlock_irqrestore(&priv->txfidx_lock, flags);

	netif_wake_queue(dev);
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC int
mpi_lan_send_reply(struct net_device *dev, LANSendReply_t *pSendRep)
{
	struct mpi_lan_priv *priv = (struct mpi_lan_priv *) dev->priv;
	MPI_ADAPTER *mpi_dev = priv->mpi_dev;
	struct sk_buff *sent;
	unsigned long flags;
	int FreeReqFrame = 0;
	u32 *pContext;
	u32 ctx;
	u8 count;

	count = pSendRep->NumberOfContexts;

	dioprintk((KERN_INFO "mpilan: send_reply: IOCStatus: %04x\n",
		 le16_to_cpu(pSendRep->IOCStatus)));

	/* Add check for Loginfo Flag in IOCStatus */

	switch (le16_to_cpu(pSendRep->IOCStatus)) {
	case MPI_IOCSTATUS_SUCCESS:
		priv->stats.tx_packets += count;
		break;

	case MPI_IOCSTATUS_LAN_CANCELED:
	case MPI_IOCSTATUS_LAN_TRANSMIT_ABORTED:
		break;

	case MPI_IOCSTATUS_INVALID_SGL:
		priv->stats.tx_errors += count;
		printk (KERN_ERR "%s: Error: invalid SGL sent to IOC\n",
			dev->name);
		return FreeReqFrame;

	default:
		priv->stats.tx_errors += count;
		break;
	}

	pContext = &pSendRep->BufferContext;

	spin_lock_irqsave(&priv->txfidx_lock, flags);
	while (count > 0) {
		ctx = le32_to_cpu(*pContext) & 0x07FFFFFF;

		sent = priv->SendCtl[ctx].skb;
		priv->stats.tx_bytes += sent->len;

		dioprintk((KERN_INFO "%s: skb %p sent.\n", __FUNCTION__, sent));

		priv->SendCtl[ctx].skb = NULL;
		pci_unmap_single(mpi_dev->pcidev, priv->SendCtl[ctx].dma,
				 priv->SendCtl[ctx].len, PCI_DMA_TODEVICE);
		dev_kfree_skb_irq(sent);

		priv->mpi_txfidx[++priv->mpi_txfidx_tail] = ctx;

		pContext++;
		count--;
	}
	spin_unlock_irqrestore(&priv->txfidx_lock, flags);

	if (!(pSendRep->MsgFlags & MPI_MSGFLAGS_CONTINUATION_REPLY))
		FreeReqFrame = 1;

	netif_wake_queue(dev);
	return FreeReqFrame;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC int
mpi_lan_sdu_send (struct sk_buff *skb, struct net_device *dev)
{
	struct mpi_lan_priv *priv = (struct mpi_lan_priv *) dev->priv;
	MPI_ADAPTER *mpi_dev = priv->mpi_dev;
	MPI_FRAME_HDR *mf;
	LANSendRequest_t *pSendReq;
	SGETransaction32_t *pTrans;
	SGESimple64_t *pSimple;
	dma_addr_t dma;
	unsigned long flags;
	int ctx;

	dioprintk((KERN_INFO "%s called, skb address = %p.\n",__FUNCTION__, skb));

	spin_lock_irqsave(&priv->txfidx_lock, flags);
	if (priv->mpi_txfidx_tail < 0) {
		netif_stop_queue(dev);
		dev_kfree_skb(skb);
		spin_unlock_irqrestore(&priv->txfidx_lock, flags);

		printk (KERN_ERR "%s: no tx context available: %u.\n",
			__FUNCTION__, priv->mpi_txfidx_tail);
		return 1;
	}

	mf = MpiGetMsgFrame(LanCtx, mpi_dev->id);
	if (mf == NULL) {
		netif_stop_queue(dev);
		dev_kfree_skb(skb);
		spin_unlock_irqrestore(&priv->txfidx_lock, flags);

		printk (KERN_ERR "%s: Unable to alloc request frame.\n",
			__FUNCTION__);
		return 1;
	}

	ctx = priv->mpi_txfidx[priv->mpi_txfidx_tail--];
	spin_unlock_irqrestore(&priv->txfidx_lock, flags);

//	dioprintk ((KERN_INFO "%s: Creating new msg frame (send).\n", dev->name));

	pSendReq = (LANSendRequest_t *) mf;

	/* Set the mac.raw pointer, since this apparently isn't getting
	 * done before we get the skb. Pull the data pointer past the mac data.
	 */
	skb->mac.raw = skb->data;
	skb_pull(skb, 12);

        dma = pci_map_single(mpi_dev->pcidev, skb->data, skb->len,
			     PCI_DMA_TODEVICE);

	priv->SendCtl[ctx].skb = skb;
	priv->SendCtl[ctx].dma = dma;
	priv->SendCtl[ctx].len = skb->len;

	/* Message Header */
	pSendReq->Function    = MPI_FUNCTION_LAN_SEND;
	pSendReq->ChainOffset = 0;
	pSendReq->MsgFlags    = 0;
	pSendReq->PortNumber  = priv->pnum;

	/* Transaction Context Element */
	pTrans = (SGETransaction32_t *) pSendReq->SG_List;

	/* No Flags, 8 bytes of Details, 32bit Context (bloody turbo replies) */
	pTrans->ContextSize   = sizeof(U32);
	pTrans->DetailsLength = 2 * sizeof(U32);
	pTrans->Flags         = 0;
	pTrans->TransactionContext[0] = cpu_to_le32(ctx);

//	dioprintk ((KERN_INFO "%s: BC = %08x, skb = %p, buff = %p.\n",
//		  dev->name, ctx, skb, skb->data));

	pTrans->TransactionDetails[0] = cpu_to_le32((0x1000          << 16) |
						    (skb->mac.raw[0] <<  8) |
						    (skb->mac.raw[1] <<  0));
	pTrans->TransactionDetails[1] = cpu_to_le32((skb->mac.raw[2] << 24) |
						    (skb->mac.raw[3] << 16) |
						    (skb->mac.raw[4] <<  8) |
						    (skb->mac.raw[5] <<  0));

	pSimple = (SGESimple64_t *) &pTrans->TransactionDetails[2];

	pSimple->FlagsLength = cpu_to_le32(
			((MPI_SGE_FLAGS_END_OF_BUFFER |
			  MPI_SGE_FLAGS_SIMPLE_ELEMENT |
			  MPI_SGE_FLAGS_64_BIT_ADDRESSING |
			  MPI_SGE_FLAGS_END_OF_LIST) << MPI_SGE_FLAGS_SHIFT) |
			skb->len);
	pSimple->Address.Low = cpu_to_le32((U32) dma);
	if (sizeof(dma_addr_t) > sizeof(U32))
		pSimple->Address.High = cpu_to_le32((U32) ((u64) dma >> 32));
	else
		pSimple->Address.High = 0;

	MpiPutMsgFrame (LanCtx, mpi_dev->id, mf);
	dev->trans_start = jiffies;

	dioprintk((KERN_INFO "%s: sending packet. FlagsLength: %08x\n",
		 dev->name, le32_to_cpu(pSimple->FlagsLength)));

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC inline void
mpi_lan_wake_post_buckets_task(struct net_device *dev)
{
	struct mpi_lan_priv *priv = dev->priv;

	if (test_and_set_bit(0, &priv->post_buckets_active) == 0) {
		queue_task(&priv->post_buckets_task, &tq_immediate);
		mark_bh(IMMEDIATE_BH);
		dioprintk (("%s: queued post_buckets task.\n", dev->name));
	}
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC inline int
mpi_lan_receive_skb(struct net_device *dev, struct sk_buff *skb)
{
	struct mpi_lan_priv *priv = dev->priv;

	skb->protocol = mpi_lan_type_trans(skb, dev);

	dioprintk((KERN_INFO "%s: Incoming packet (%d bytes) " 
		 "delivered to upper level.\n", dev->name, skb->len));

	priv->stats.rx_bytes += skb->len;
	priv->stats.rx_packets++;

	skb->dev = dev;
	netif_rx(skb);

	dioprintk(("mpilan/receive_skb: %d buckets remaining.\n",
		 atomic_read(&priv->buckets_out)));

	if (atomic_read(&priv->buckets_out) < priv->bucketthresh)
		mpi_lan_wake_post_buckets_task(dev);

	dioprintk ((KERN_INFO "mpilan/receive_post_reply: %d buckets "
		  "remaining, %d received back since sod.\n",
		  atomic_read(&priv->buckets_out), priv->total_received));

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
//STATIC inline int
STATIC int
mpi_lan_receive_post_turbo(struct net_device *dev, u32 tmsg)
{
	struct mpi_lan_priv *priv = dev->priv;
	MPI_ADAPTER *mpi_dev = priv->mpi_dev;
	struct sk_buff *skb, *old_skb;
	unsigned long flags;
	u32 ctx, len;

	ctx = (tmsg >> 16) & 0x7FF;
	skb = priv->RcvCtl[ctx].skb;

	len = tmsg & 0x0000FFFF;

	if (len < MPI_LAN_RX_COPYBREAK) {
		old_skb = skb;

		skb = (struct sk_buff *)dev_alloc_skb(len);
		if (!skb) {
			printk (KERN_ERR "%s: Can't allocate skb - %s@%d.\n",
				dev->name, __FILE__, __LINE__);
			return -ENOMEM;
		}

		pci_dma_sync_single(mpi_dev->pcidev, priv->RcvCtl[ctx].dma,
				    priv->RcvCtl[ctx].len, PCI_DMA_FROMDEVICE);

		memcpy(skb_put(skb, len), old_skb->data, len);

		goto out;
	}

	skb_put(skb, len);

	priv->RcvCtl[ctx].skb = NULL;

	pci_unmap_single(mpi_dev->pcidev, priv->RcvCtl[ctx].dma,
			 priv->RcvCtl[ctx].len, PCI_DMA_FROMDEVICE);

out:
	spin_lock_irqsave(&priv->rxfidx_lock, flags);
	priv->mpi_rxfidx[++priv->mpi_rxfidx_tail] = ctx;
	spin_unlock_irqrestore(&priv->rxfidx_lock, flags);

	atomic_dec(&priv->buckets_out);
	priv->total_received++;

	return mpi_lan_receive_skb(dev, skb);
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC int
mpi_lan_receive_post_free(struct net_device *dev,
			  LANReceivePostReply_t *pRecvRep)
{
	struct mpi_lan_priv *priv = dev->priv;
	MPI_ADAPTER *mpi_dev = priv->mpi_dev;
	unsigned long flags;
	struct sk_buff *skb;
	u32 ctx;
	u8 count;
	int i;

	count = pRecvRep->NumberOfContexts;

/**/	dprintk ((KERN_INFO "mpilan/receive_post_reply: "
		  "IOC returned %d buckets, freeing.\n", count));

	spin_lock_irqsave(&priv->rxfidx_lock, flags);
	for (i = 0; i < count; i++) {
		ctx = le32_to_cpu(pRecvRep->BucketContext[i]);

		skb = priv->RcvCtl[ctx].skb;

//		dprintk (("mpilan/rpr:(2) priv = %p, buckets_out addr = %p, ",
//			  priv, &(priv->buckets_out)));
//		dprintk(("dev name = %s.\n", dev->name));
//		dprintk(("mpilan/receive_post_reply: (2) TC +3.\n"));

		priv->RcvCtl[ctx].skb = NULL;
		pci_unmap_single(mpi_dev->pcidev, priv->RcvCtl[ctx].dma,
				 priv->RcvCtl[ctx].len, PCI_DMA_FROMDEVICE);
		dev_kfree_skb_any(skb);

		priv->mpi_rxfidx[++priv->mpi_rxfidx_tail] = ctx;
	}
	spin_unlock_irqrestore(&priv->rxfidx_lock, flags);

	atomic_sub(count, &priv->buckets_out);

//	for (i = 0; i < priv->max_buckets_out; i++)
//		if (priv->RcvCtl[i].skb != NULL)
//			dprintk ((KERN_INFO "mpilan/rpr: bucket %03x "
//				  "is still out.\n", i));

/*	dprintk ((KERN_INFO "mpilan/receive_post_reply: freed %d buckets.\n",
		  count));
*/
/**/	dprintk ((KERN_INFO "mpilan/receive_post_reply: %d buckets "
/**/		  "remaining, %d received back since sod.\n",
/**/		  atomic_read(&priv->buckets_out), priv->total_received));
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC int
mpi_lan_receive_post_reply(struct net_device *dev,
			   LANReceivePostReply_t *pRecvRep)
{
	struct mpi_lan_priv *priv = dev->priv;
	MPI_ADAPTER *mpi_dev = priv->mpi_dev;
	struct sk_buff *skb, *old_skb;
	unsigned long flags;
	u32 len, ctx;
	u32 offset, remaining;
	u8 count;
	int i, l;

	dioprintk((KERN_INFO "mpilan: mpi_lan_receive_post_reply called.\n"));
	dioprintk((KERN_INFO "mpilan: receive_post_reply: IOCStatus: %04x\n",
		 le16_to_cpu(pRecvRep->IOCStatus)));

	if (le16_to_cpu(pRecvRep->IOCStatus) & MPI_IOCSTATUS_LAN_CANCELED)
		return mpi_lan_receive_post_free(dev, pRecvRep);

	len = le32_to_cpu(pRecvRep->PacketLength);
	if (len == 0) {
		printk (KERN_ERR "mpi/rpr: Got a non-Turbo "
			"ReceivePostReply w/ PacketLength zero\n");
		printk (KERN_ERR "mpi/rpr: MsgFlags = %02x, IOCStatus = %04x\n",
			pRecvRep->MsgFlags, le16_to_cpu(pRecvRep->IOCStatus));
	 	if (le16_to_cpu(pRecvRep->IOCStatus) & MPI_IOCSTATUS_FLAG_LOG_INFO_AVAILABLE)
			printk (KERN_ERR "mpi/rpr: LogInfo = %08x.\n",
				le32_to_cpu(pRecvRep->IOCLogInfo));
		return -1;
	}

	ctx    = le32_to_cpu(pRecvRep->BucketContext[0]);
	count  = pRecvRep->NumberOfContexts;
	skb    = priv->RcvCtl[ctx].skb;

	offset = le32_to_cpu(pRecvRep->PacketOffset);
//	if (offset != 0)
//		printk (KERN_INFO "mpi/rpr: Got a ReceivePostReply "
//			"w/ PacketOffset %u.\n", offset);

	dioprintk((KERN_INFO "mpilan/rpr: offset = %d, len = %d\n", offset, len));

	if (count > 1) {
		int szrem = len;

/*		dioprintk ((KERN_INFO "Multiple buckets returned "
			  "for single packet, concatenating.\n"));
*/
		skb = (struct sk_buff *)dev_alloc_skb(len);
		if (!skb) {
			printk (KERN_ERR "%s: Can't allocate skb - %s@%d.\n",
				dev->name, __FILE__, __LINE__);
			return -ENOMEM;
		}

		spin_lock_irqsave(&priv->rxfidx_lock, flags);
		for (i = 0; i < count; i++) {

			ctx = le32_to_cpu(pRecvRep->BucketContext[i]);
			old_skb = priv->RcvCtl[ctx].skb;

			l = priv->RcvCtl[ctx].len;
			if (szrem < l)
				l = szrem;

//			dioprintk((KERN_INFO "Buckets %d: len %u\n", i, l));

			pci_dma_sync_single(mpi_dev->pcidev,
					    priv->RcvCtl[ctx].dma,
					    priv->RcvCtl[ctx].len,
					    PCI_DMA_FROMDEVICE);
			memcpy(skb_put(skb, l), old_skb->data, l);

			priv->mpi_rxfidx[++priv->mpi_rxfidx_tail] = ctx;
			szrem -= l;
		}
		spin_unlock_irqrestore(&priv->rxfidx_lock, flags);

	} else if (len < MPI_LAN_RX_COPYBREAK) {

		old_skb = skb;

		skb = (struct sk_buff *)dev_alloc_skb(len);
		if (!skb) {
			printk (KERN_ERR "%s: Can't allocate skb - %s@%d.\n",
				dev->name, __FILE__, __LINE__);
			return -ENOMEM;
		}

		pci_dma_sync_single(mpi_dev->pcidev, priv->RcvCtl[ctx].dma,
				    priv->RcvCtl[ctx].len, PCI_DMA_FROMDEVICE);

		memcpy(skb_put(skb, len), old_skb->data, len);

		spin_lock_irqsave(&priv->rxfidx_lock, flags);
		priv->mpi_rxfidx[++priv->mpi_rxfidx_tail] = ctx;
		spin_unlock_irqrestore(&priv->rxfidx_lock, flags);

	} else {
		spin_lock_irqsave(&priv->rxfidx_lock, flags);

		priv->RcvCtl[ctx].skb = NULL;

		pci_unmap_single(mpi_dev->pcidev, priv->RcvCtl[ctx].dma,
				 priv->RcvCtl[ctx].len, PCI_DMA_FROMDEVICE);
		priv->RcvCtl[ctx].dma = 0;

		priv->mpi_rxfidx[++priv->mpi_rxfidx_tail] = ctx;
		spin_unlock_irqrestore(&priv->rxfidx_lock, flags);

		skb_put(skb,len);
	}

	atomic_sub(count, &priv->buckets_out);
	priv->total_received += count;

	if (priv->mpi_rxfidx_tail >= MPI_LAN_MAX_BUCKETS_OUT) {
		printk (KERN_ERR "mpilan/rpr: Yoohoo! "
			"priv->mpi_rxfidx_tail = %d, "
			"MPI_LAN_MAX_BUCKETS_OUT = %d.\n",
			priv->mpi_rxfidx_tail, MPI_LAN_MAX_BUCKETS_OUT);

		panic("Damn it Jim! I'm a doctor, not a programmer! "
		      "Oh, wait a sec, I am a programmer. "
		      "And, who's Jim?!?!\n"
		      "Arrgghh! We've done it again!.\n");
	}

//	remaining = le32_to_cpu(pRecvRep->BucketsRemaining);
//	if (remaining < priv->bucketthresh)
//		mpi_lan_wake_post_buckets_task(dev);

//	if (remaining == 0)
//		printk (KERN_WARNING "%s: DDM out of buckets "
//			"(priv->buckets_out = %d)!\n",
//			dev->name, atomic_read(&priv->buckets_out));
//	else
//		printk (KERN_INFO "%s: DDM says %d buckets left "
//			"(priv->buckets_out = %d)!\n",
//			dev->name, remaining, atomic_read(&priv->buckets_out));

	return mpi_lan_receive_skb(dev, skb);
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/* Simple SGE's only at the moment */

STATIC void
mpi_lan_post_receive_buckets(void *dev_id)
{
	struct net_device *dev = dev_id;
	struct mpi_lan_priv *priv = dev->priv;
	MPI_ADAPTER *mpi_dev = priv->mpi_dev;
	MPI_FRAME_HDR *mf;
	LANReceivePostRequest_t *pRecvReq;
	SGETransaction32_t *pTrans;
	SGESimple64_t *pSimple;
	struct sk_buff *skb;
	dma_addr_t dma;
	u32 curr, buckets, count, max;
	u32 len = (dev->mtu + dev->hard_header_len + 4);
	unsigned long flags;
	int i;

	curr = atomic_read(&priv->buckets_out);
	buckets = (priv->max_buckets_out - curr);

	dioprintk((KERN_INFO "%s: Start. buckets = %u, buckets out = %u.\n",
	__FUNCTION__, buckets, curr));

	max = (mpi_dev->req_sz - MPI_LAN_RECEIVE_POST_REQUEST_SIZE) /
			(MPI_LAN_TRANSACTION32_SIZE + sizeof(SGESimple64_t));

	while (buckets) {
		mf = MpiGetMsgFrame(LanCtx, mpi_dev->id);
		if (mf == NULL) {
			printk (KERN_ERR "%s: Unable to alloc request frame.\n",
				__FUNCTION__);
			dioprintk((KERN_ERR "%s: %u buckets remaining.\n",
				 __FUNCTION__, buckets));
			goto out;
		}
		pRecvReq = (LANReceivePostRequest_t *) mf;

		count = buckets;
		if (count > max)
			count = max;

		pRecvReq->Function    = MPI_FUNCTION_LAN_RECEIVE;
		pRecvReq->ChainOffset = 0;
		pRecvReq->MsgFlags    = 0;
		pRecvReq->PortNumber  = priv->pnum;

		pTrans = (SGETransaction32_t *) pRecvReq->SG_List;
		pSimple = NULL;

		for (i = 0; i < count; i++) {
			int ctx;

			spin_lock_irqsave(&priv->rxfidx_lock, flags);
			if (priv->mpi_rxfidx_tail < 0) {
				printk (KERN_ERR "%s: Can't alloc context.\n",
					__FUNCTION__);
				spin_unlock_irqrestore(&priv->rxfidx_lock,
						       flags);
				break;
			}

			ctx = priv->mpi_rxfidx[priv->mpi_rxfidx_tail--];

			skb = priv->RcvCtl[ctx].skb;
			if (skb && (priv->RcvCtl[ctx].len != len)) {
				pci_unmap_single(mpi_dev->pcidev,
						 priv->RcvCtl[ctx].dma,
						 priv->RcvCtl[ctx].len,
						 PCI_DMA_FROMDEVICE);
				dev_kfree_skb(priv->RcvCtl[ctx].skb);
				skb = priv->RcvCtl[ctx].skb = NULL;
			}

			if (skb == NULL) {
				skb = dev_alloc_skb(len);
				if (skb == NULL) {
/**/					printk (KERN_WARNING
/**/						"%s: Can't alloc skb.\n",
/**/						__FUNCTION__);
					priv->mpi_rxfidx[++priv->mpi_rxfidx_tail] = ctx;
					spin_unlock_irqrestore(&priv->rxfidx_lock, flags);
					break;
				}

		    		dma = pci_map_single(mpi_dev->pcidev, skb->data,
						     len, PCI_DMA_FROMDEVICE);

				priv->RcvCtl[ctx].skb = skb;
				priv->RcvCtl[ctx].dma = dma;
				priv->RcvCtl[ctx].len = len;
			}

			spin_unlock_irqrestore(&priv->rxfidx_lock, flags);

			pTrans->ContextSize   = sizeof(U32);
			pTrans->DetailsLength = 0;
			pTrans->Flags         = 0;
			pTrans->TransactionContext[0] = cpu_to_le32(ctx);

			pSimple = (SGESimple64_t *) pTrans->TransactionDetails;

			pSimple->FlagsLength = cpu_to_le32(
				((MPI_SGE_FLAGS_END_OF_BUFFER |
				  MPI_SGE_FLAGS_SIMPLE_ELEMENT |
				  MPI_SGE_FLAGS_64_BIT_ADDRESSING) << MPI_SGE_FLAGS_SHIFT) | len);
			pSimple->Address.Low = cpu_to_le32((U32) priv->RcvCtl[ctx].dma);
			if (sizeof(dma_addr_t) > sizeof(U32))
				pSimple->Address.High = cpu_to_le32((U32) ((u64) priv->RcvCtl[ctx].dma >> 32));
			else
				pSimple->Address.High = 0;

			pTrans = (SGETransaction32_t *) (pSimple + 1);
		}

		if (pSimple == NULL) {
/**/			printk (KERN_WARNING "%s: No buckets posted.\n",
/**/				__FUNCTION__);
			MpiFreeMsgFrame(LanCtx, mpi_dev->id, mf);
			goto out;
		}

		pSimple->FlagsLength |= cpu_to_le32(MPI_SGE_FLAGS_END_OF_LIST << MPI_SGE_FLAGS_SHIFT);

		pRecvReq->BucketCount = cpu_to_le32(i);

/*	printk(KERN_INFO "mpilan: posting buckets\n   ");
 *	for (i = 0; i < j + 2; i ++)
 *	    printk (" %08x", le32_to_cpu(msg[i]));
 *	printk ("\n");
 */

		MpiPutMsgFrame(LanCtx, mpi_dev->id, mf);

		priv->total_posted += i;
		buckets -= i;
		atomic_add(i, &priv->buckets_out);
	}

out:
	dioprintk ((KERN_INFO "%s: End. buckets = %u, priv->buckets_out = %u.\n",
		  __FUNCTION__, buckets, atomic_read(&priv->buckets_out)));
	dioprintk ((KERN_INFO "%s: Posted %u buckets and received %u back.\n",
	__FUNCTION__, priv->total_posted, priv->total_received)); 

	clear_bit(0, &priv->post_buckets_active);
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
struct net_device *
mpi_register_lan_device (MPI_ADAPTER *mpi_dev, int pnum)
{
	struct net_device *dev = NULL;
	struct mpi_lan_priv *priv = NULL;
	u8 HWaddr[FC_ALEN], *a;

	dev = init_fcdev(NULL, sizeof(struct mpi_lan_priv));
	if (!dev)
		return (NULL);
	dev->mtu = MPI_LAN_MTU;

	priv = (struct mpi_lan_priv *) dev->priv;

	priv->mpi_dev = mpi_dev;
	priv->pnum = pnum;

	memset(&priv->post_buckets_task, 0, sizeof(struct tq_struct));
	priv->post_buckets_task.routine = mpi_lan_post_receive_buckets;
	priv->post_buckets_task.data = dev;
	priv->post_buckets_active = 0;

	printk (KERN_INFO "%s@%d: bucketlen = %d.\n",
		__FILE__, __LINE__, dev->mtu + dev->hard_header_len + 4);

	atomic_set(&priv->buckets_out, 0);
	priv->total_posted = 0;
	priv->total_received = 0;
	priv->max_buckets_out = max_buckets_out;
	if (mpi_dev->pfacts0.MaxLanBuckets < max_buckets_out)
		priv->max_buckets_out = mpi_dev->pfacts0.MaxLanBuckets;

	dprintk ((KERN_INFO "%s@%d: MaxLanBuckets = %d, max_buckets_out = %d, "
		  "priv->max_buckets_out = %d.\n",
		  __FILE__, __LINE__, mpi_dev->pfacts0.MaxLanBuckets,
		  max_buckets_out, priv->max_buckets_out));

	priv->bucketthresh = priv->max_buckets_out * 2 / 3;
	priv->txfidx_lock = SPIN_LOCK_UNLOCKED;
	priv->rxfidx_lock = SPIN_LOCK_UNLOCKED;

	memset(&priv->stats, 0, sizeof(priv->stats));

	/*  Grab pre-fetched LANPage1 stuff. :-) */
	a = (u8 *) &mpi_dev->lan_cnfg_page1.HardwareAddressLow;

	HWaddr[0] = a[5];
	HWaddr[1] = a[4];
	HWaddr[2] = a[3];
	HWaddr[3] = a[2];
	HWaddr[4] = a[1];
	HWaddr[5] = a[0];

	printk (KERN_INFO "%s: hwaddr = %02X:%02X:%02X:%02X:%02X:%02X\n",
	        dev->name, HWaddr[0], HWaddr[1], HWaddr[2],
		HWaddr[3], HWaddr[4], HWaddr[5]);

	dev->addr_len = FC_ALEN;
	memcpy(dev->dev_addr, HWaddr, FC_ALEN);
	memset(dev->broadcast, 0xff, FC_ALEN);

	/* The Tx queue is 127 deep on the 909.
	 * Give ourselves some breathing room.
	 */
	priv->tx_max_out = (tx_max_out_p <= MPI_TX_MAX_OUT_LIM) ?
			    tx_max_out_p : MPI_TX_MAX_OUT_LIM;
	printk (KERN_INFO "%s: Max TX Outstanding = %d.\n",
		dev->name, priv->tx_max_out);

	dev->open = mpi_lan_open;
	dev->stop = mpi_lan_close;
	dev->get_stats = mpi_lan_get_stats;
	dev->set_multicast_list = NULL;
	dev->change_mtu = mpi_lan_change_mtu;
	dev->hard_start_xmit = mpi_lan_sdu_send;

/* Not in 2.3.42. Need 2.3.45+ */
	dev->tx_timeout = mpi_lan_tx_timeout;
	dev->watchdog_timeo = MPI_LAN_TX_TIMEOUT;

	printk (KERN_INFO "mpilan: Finished registering dev "
		"and setting initial values.\n");

	return dev;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
int __init
mpi_lan_init (void)
{
	struct net_device *dev;
	MPI_ADAPTER *curadapter;
	int i = 0, j;
	
	show_mpimod_ver(LANAME, LANVER);

	if ((LanCtx = mpi_register(lan_reply)) < 0) {
		printk (KERN_ERR "mpilan: Failed to register with MPI drivercore.\n");
		return -EBUSY;
	}

	/* Set the callback index to be used by driver core for turbo replies */
	LanCIdx = LanCtx;

	dprintk((KERN_INFO "mpilan: assigned context of %d.\n", LanCtx));

	for (j = 0; j < myMPI_MAX_ADAPTERS; j++) {
		mpi_landev[j] = NULL;
	}
	j = 0;

	curadapter = mpi_adapter_find_first();
	while (curadapter != NULL) {
		for (i = 0; i < curadapter->facts0.NumberOfPorts; i++) {
			if (curadapter->pfacts0.ProtocolFlags & MPI_PORTFACTS_PROTOCOL_LAN) {
				dev = mpi_register_lan_device (curadapter, i);
				if (dev != NULL) {
					printk (KERN_INFO "%s: Fusion MPT LAN device registered - ioc %d\n",
						dev->name, curadapter->id);
					printk (KERN_INFO "mpilan: adapter %d, port %d, PROTOCOL = %02x.\n", curadapter->id, curadapter->facts0.NumberOfPorts, curadapter->pfacts0.ProtocolFlags);
					mpi_landev[j] = dev;
					dprintk((KERN_INFO "mpilan/init: dev address = %p, mpi_landev[%d] = %p.\n", dev, j,  mpi_landev[j]));
					j++;
				} else {
					printk (KERN_ERR "mpilan: Unable to register ioc %d as a LAN device.\n",
						curadapter->id);
				}
			}
			else
			{
			printk (KERN_INFO "mpilan: adapter %d, port %d, PROTOCOL = %02x.\n", curadapter->id, curadapter->facts0.NumberOfPorts, curadapter->pfacts0.ProtocolFlags);
			}
		}
		curadapter = mpi_adapter_find_next(curadapter);
	}

	return 0;
}

#ifdef MODULE
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
void
cleanup_module(void)
{
	int i;

	for (i = 0; mpi_landev[i] != NULL; i++) {
		struct net_device *dev = mpi_landev[i];

		unregister_fcdev(dev);
		mpi_landev[i] = (struct net_device *) 0xdeadbeef; /* Debug */
		printk (KERN_INFO "%s: Fusion MPT LAN device unregistered.\n", dev->name);
	}

	if (LanCtx >= 0) {
		mpi_deregister(LanCtx);
		LanCtx = -1;
		LanCIdx = 0;
	}

	/* deregister any send/receive handler structs. I2Oism? */
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

MODULE_PARM(tx_max_out_p, "i");
MODULE_PARM(max_buckets_out, "i"); // Debug stuff. FIXME!

#endif

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
STATIC unsigned short
mpi_lan_type_trans(struct sk_buff *skb, struct net_device *dev) 
{
	struct mpi_lan_ohdr *fch = (struct mpi_lan_ohdr *)skb->data;
	struct fcllc *fcllc;

	skb->mac.raw = skb->data;
	skb_pull(skb, sizeof(struct mpi_lan_ohdr));

	if (fch->dtype == htons(0xffff)) {
		u32 *p = (u32 *) fch;

		swab32s(p + 0);
		swab32s(p + 1);
		swab32s(p + 2);
		swab32s(p + 3);

		printk (KERN_INFO "%s: BROADCAST swap F/W bug, please update "
			"sender (%02x:%02x:%02x:%02x:%02x:%02x).\n",
			dev->name, fch->saddr[0], fch->saddr[1], fch->saddr[2],
			fch->saddr[3], fch->saddr[4], fch->saddr[5]);
	}

//	dprintk((KERN_INFO "Fc Optional Header:"));
//	dprintk((" %04x", ntohs(fch->dtype)));
//	dprintk((" %02x:%02x:%02x:%02x:%02x:%02x",
//		 fch->daddr[0], fch->daddr[1], fch->daddr[2],
//		 fch->daddr[3], fch->daddr[4], fch->daddr[5]));
//	dprintk((" %04x", ntohs(fch->stype)));
//	dprintk((" %02x:%02x:%02x:%02x:%02x:%02x",
//		 fch->saddr[0], fch->saddr[1], fch->saddr[2],
//		 fch->saddr[3], fch->saddr[4], fch->saddr[5]));
//	dprintk(("\n"));

	if (*fch->daddr & 1) {
//		dprintk((KERN_DEBUG "mpi_lan_type_trans: "
//			 "checking for [broad|multi]cast.\n"));
		if (!memcmp(fch->daddr, dev->broadcast, FC_ALEN)) {
			skb->pkt_type = PACKET_BROADCAST;
//			dprintk ((KERN_DEBUG "mpi_lan_type_trans: "
//				  "set pkt_type = BROADCAST.\n"));
		} else {
			skb->pkt_type = PACKET_MULTICAST;
//			dprintk ((KERN_DEBUG "mpi_lan_type_trans: "
//				  "set pkt_type = MULTICAST.\n"));
		}
	} else {
		if (memcmp(fch->daddr, dev->dev_addr, FC_ALEN)) {
			skb->pkt_type = PACKET_OTHERHOST;
//			dprintk ((KERN_DEBUG "mpi_lan_type_trans: "
//				  "set pkt_type = OTHERHOST.\n"));
		} else {
			skb->pkt_type = PACKET_HOST;
//			dprintk ((KERN_DEBUG "mpi_lan_type_trans: "
//				  "set pkt_type = HOST.\n"));
		}
	}

	/* Strip the SNAP header from ARP packets since we don't 
	 * pass them through to the 802.2/SNAP layers.
	 */
	fcllc = (struct fcllc *)skb->data;

	if (fcllc->dsap == EXTENDED_SAP &&
		(fcllc->ethertype == htons(ETH_P_IP) ||
		 fcllc->ethertype == htons(ETH_P_ARP))) {
		skb_pull(skb, sizeof(struct fcllc));
//		dprintk ((KERN_DEBUG "mpi_lan_type_trans: "
//			  "returning ethertype = %04x "
//			  "(IP = %04x, ARP = %04x).\n",
//			  ntohs(fcllc->ethertype), ETH_P_IP, ETH_P_ARP));
		return fcllc->ethertype;
	}

//	dprintk ((KERN_DEBUG "mpi_lan_type_trans: returning ETH_P_802_2.\n"));
	return htons(ETH_P_802_2);
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
