/*
 *	rrw.c - MPI raw SCSI read/write generator for linux+mpi
 */
/*==========================================================================*/

#include <stdio.h>
#include <stdlib.h>		/* for atoi() */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>		/* for gettimeofday() */
#include <unistd.h>		/* for gettimeofday(), getopt() */
#include <sys/ioctl.h>
#include "mpi.h"		/* everything important to us */

/*==========================================================================*/

#define TARG_DEF    0
//#define ITERS       10000
#define ITERS       1
//#define ITERS_LOTS  100000
#define ITERS_LOTS  10
#define IOSIZ_DEF   16
#define QD_DEF      32
#define CACHE_DEF   64
#define VERB_DEF    1

static unsigned int ctl_type = MPIRWPERF;
static int verbose = VERB_DEF;

static int parse_args(int argc, char *argv[], const char *myname, struct mpi_raw_r_w *pi);


/*==========================================================================*/
static int
usage(const char *myname, int errcode)
{

fprintf(stderr, "%ssage:  %s  [options]\n\n",
	(errcode == -1) ? "U" : "Correct u",
	myname);

fprintf(stderr, "  Options:\n");
fprintf(stderr, "    -i N     ioc#      (default=0, range={0-?})\n");
fprintf(stderr, "    -t N     target#   (default=%d, N=99 --> BadTargetTest!)\n",
	TARG_DEF);
fprintf(stderr, "    -n N     NumIters  (default=%d)\n",
	ITERS);
fprintf(stderr, "    -b N[km] IoSize    In BLOCKS! (default=%d == %dkB; k/m suffix --> kB/mB)\n",
	IOSIZ_DEF, IOSIZ_DEF/2);
//fprintf(stderr, "                                                (N=0 --> 0-Byte READ test!)\n");
fprintf(stderr, "    -q N     Qdepth    (default=%d)\n",
	QD_DEF);
fprintf(stderr, "    -c N[km] CacheSz   (in Kb!  default=0kB == OFF!)\n");
fprintf(stderr, "             --------------------------------------------------------------\n");
fprintf(stderr, "    -w P     %% WRITEs  RD/WR mix    (DEFAULT = 100%% pure READs!)\n");
fprintf(stderr, "                                    (P=100 --> 100%% pure WRITEs)\n");
fprintf(stderr, "    -m P     %% SEQ     SEQ/RAND mix (DEFAULT = 100%% pure sequential!)\n");
fprintf(stderr, "                                    (P=0 ----> 100%% pure random)\n");
fprintf(stderr, "    -r P     %% Range   (use  0-100%% of total capacity)\n");
fprintf(stderr, "    -k P     %% Skip    (skip 0-99%%  of total capacity)\n");
fprintf(stderr, "             --------------------------------------------------------------\n");
fprintf(stderr, "    -C       CHECK/status request [EXP - Check klogd output.]\n");
fprintf(stderr, "    -R       RESET! request       [EXP - Use with CAUTION!!!]\n");
fprintf(stderr, "             --------------------------------------------------------------\n");
fprintf(stderr, "    -v       verbose   (flag/toggle, default=%s)\n",
	VERB_DEF ? "ON!" : "off");
fprintf(stderr, "    -h       Help      (this text)\n\n");

	return errcode;
}


/*==========================================================================*/
int
main (int argc, char *argv[])
{
		struct mpi_raw_r_w PerfInfo;
	int mpifd;
	int ioctlrv;
		struct timeval tv1, tv2;
		long double IOPS;
		long double delta_time;
	int errnum = 1;
		char *myname;


	if ( (myname = strrchr(argv[0], '/')) == NULL )
		myname = argv[0];
	else
	{
		myname++;
		argv[0] = myname;	// Yikes!  Is that legal?
	}

/* build */
		/* fill out PerfInfo - defaults */
	PerfInfo.iocport = 0;
	PerfInfo.btl = 0;		// 0:0:0 is usually the first SCSI device
	PerfInfo.nblks = IOSIZ_DEF;
	PerfInfo.qdepth = QD_DEF;
	PerfInfo.iters = ITERS;
	PerfInfo.skip = 0;
	PerfInfo.range = 0;		// range of disk
	PerfInfo.rdwr = 0;		// 0=pure READs
	PerfInfo.seqran = 0;		// 0=pure SEQuential
	PerfInfo.cache_sz = 0;		// Cache OFF!

/* cmd-line overrides... */
	if ( (ioctlrv = parse_args(argc, argv, myname, &PerfInfo)) != 0 )
		return (ioctlrv == -1) ? 0 : ioctlrv;

	if (PerfInfo.btl == -1)
		PerfInfo.nblks = 0;

//	if (PerfInfo.nblks == 0 && PerfInfo.iters == ITERS)
//	{
//		fprintf(stderr, "  C'mon - bumping iters from %d to %d\n", ITERS, ITERS_LOTS);
//		PerfInfo.iters = ITERS_LOTS;
//	}

/* say what we're going to do */
	if (verbose)
	{
		char	 foo[32];
		double	 md;
		int	 m;
		int	 h;

		fprintf(stderr, "  Info: using ioc=%d\n", PerfInfo.iocport);
		fprintf(stderr, "  Info: using b:t:l=%06xh\n", PerfInfo.btl);
		fprintf(stderr, "  Info: using iters=%d\n", PerfInfo.iters);

	    h = (PerfInfo.nblks % 2) * 5;
	    sprintf(foo, "%d%skB", (int)PerfInfo.nblks/2, h ? ".5" : "");
	    if ( PerfInfo.nblks >= 2048 )	// Mb check
	    {
	      md = PerfInfo.nblks / 2048.0;
	      m = md;
	      sprintf(foo+strlen(foo), " (%d.%d Mb!)", m, (int)((md-m+0.0005)*1000.0));
	    }

		fprintf(stderr, "  Info: using nblks=%d==%s\n", (int)PerfInfo.nblks, foo);
		fprintf(stderr, "  Info: using qdepth=%d\n", (int)PerfInfo.qdepth);
		fprintf(stderr, "  Info: using skip=%d%%\n", (int)PerfInfo.skip);
		fprintf(stderr, "  Info: using cache_sz=%dkB\n", PerfInfo.cache_sz);
	}

/* DoIt! */
/* open */
		/* /dev/mpictl *must* open first */
	if ((mpifd = open ("/dev/mpictl", O_RDWR)) == -1)
	{
		fprintf(stderr, "error opening /dev/mpictl: %s \n", strerror(errno));
		return errnum++;
	}

/* time1 */
	if ( gettimeofday(&tv1, NULL) != 0 )
	{
		fprintf(stderr, "  Error: gettimeofday(1) call failed: %s\n", strerror(errno));
		return errnum++;
	}

/* ioctl send!*/
	ioctlrv = ioctl(mpifd, ctl_type, &PerfInfo);

/* time2 */
	if ( gettimeofday(&tv2, NULL) != 0 )
	{
		fprintf(stderr, "  Error: gettimeofday(2) call failed: %s\n", strerror(errno));
		return errnum++;
	}

/* gonk results */
	delta_time = tv2.tv_sec - tv1.tv_sec
			 + (tv2.tv_usec - tv1.tv_usec)/(long double)1000000.0;

	//  avoid divide-by-0!
	if ( delta_time == 0.0 )
		delta_time = 0.000001;

// bug!
	if ( delta_time < 0.0 )
		delta_time = (tv2.tv_sec + 24.0*60.0*60.0) - tv1.tv_sec
				 + (tv2.tv_usec - tv1.tv_usec)/(long double)1000000.0;

	IOPS = (long double)PerfInfo.iters / delta_time;

/* report results */
	if (ioctlrv && (ctl_type != MPIRWPERF_CHK))
	{
		fprintf(stderr, "8-(  rv=%d\n", ioctlrv);
		fprintf(stderr, "  Error: %s\n", strerror(errno));
		fprintf(stderr, "  Error: %s\n", strerror(ioctlrv));
	}
	else
	{
		if (ctl_type != MPIRWPERF)
			verbose = 0;

		if (verbose)
		{
			fprintf(stderr, "    Success!   8-)\n");
			fprintf(stderr, "  Result: %d iterations / %LG sec = %LG IOPS\n",
					PerfInfo.iters, delta_time, IOPS);

			if (PerfInfo.nblks)
				fprintf(stderr, "  Info: Throughput = %LG Kb/s\n",
						IOPS * PerfInfo.nblks * 0.5);
		}

		if (ctl_type == MPIRWPERF)
			printf("%d.%d,%d,%LG,%d,%d\n",
				(int)PerfInfo.nblks/2, (int)(PerfInfo.nblks%2)*5,
				PerfInfo.iters,
				delta_time,
				(int)(IOPS+0.5),
				(int)(IOPS*PerfInfo.nblks*0.5+0.5));

		else if (ctl_type == MPIRWPERF_CHK)
			printf("%d\n", ioctlrv);
		else if (ctl_type == MPIRWPERF_RESET)
			printf("RESET_DONE!\n");
	}

	return ioctlrv;
}


/*==========================================================================*/
static int
parse_args(int argc, char *argv[], const char *myname, struct mpi_raw_r_w *pi)
{
		extern char *optarg;
		extern int optind, optopt;
	int c;


	while ((c = getopt(argc, argv, "i:t:n:b:q:c:w:m:r:k:CRhvs:S")) != -1)
	switch (c)
	{
		case 'i':					// ioc#
			pi->iocport = *optarg - '0';
			if (pi->iocport < 0 || pi->iocport > 1)
			{
				fprintf(stderr, "  Error: use ioc value of 0 or 1\n");
				return 'i'-'a'+1;
			}
                        pi->iocport = pi->iocport << 8;
			break;

		case 't':					// target#
			pi->btl = atoi(optarg);
                    // Oops!
			pi->btl <<= 8;
			break;

		case 'n':
// temporary! (?)
			if ( (pi->iters = atoi(optarg)) == 0 )
			{
				fprintf(stderr, "  Warning: Resetting iters to 1!\n");
				pi->iters = 1;
			}
			break;

		case 'b':					// per IO block size
		case 's':					// hidden alias for -b! (-size)
			pi->nblks = atoi(optarg);
			if (optarg[strlen(optarg)-1] == 'k')
				pi->nblks *= 2;
			if (optarg[strlen(optarg)-1] == 'm')
				pi->nblks *= 2048;

// limit #1
//  Linux will only let us kmalloc() up to 128 Kb!
//			if ( (pi->nblks == 0) || (pi->nblks > 200) )
// Minimal chain SG support added.
// We max out at 10 SGE's in the request frame!
//			if (pi->nblks > 2048)		// ~(512 Kb - 1 Mb) max! (?)
// limit #2
// Need chain SG support!  We max out at 75 SGE's!
			if (pi->nblks > 32768)		// ?? Mb max ??
			{
				fprintf(stderr, "  Warning: Resetting nblks from %d to %d!\n",
						pi->nblks, pi->nblks/2);
				pi->nblks /= 2;
			}
			break;

		case 'q':					// Qdepth
			pi->qdepth = atoi(optarg);
			if ( pi->qdepth == 0 )
			{
				fprintf(stderr, "  Warning: Resetting qdepth from 0 to 1!\n");
				pi->qdepth = 1;
			}
			else if ( pi->qdepth > 256 )
			{
				fprintf(stderr, "  Warning: Resetting qdepth from %d to 256!\n",
						pi->qdepth);
				pi->qdepth = 256;
			}
			break;

		case 'c':					// Cache size (Kb)
			pi->cache_sz = atoi(optarg);
			if (optarg[strlen(optarg)-1] == 'k')
				pi->cache_sz *= 1;
			if (optarg[strlen(optarg)-1] == 'm')
				pi->cache_sz *= 1024;
			break;

		case 'k':					// Skip%
			pi->skip = atoi(optarg);
			if ( pi->skip > 100 )
			{
				fprintf(stderr, "  Warning: Resetting skip from %d to 99%!\n",
						pi->skip);
				pi->skip = 99;
			}

		case 'w':					// WRITE%
		case 'm':					// Mix%
		case 'r':					// Range%
			fprintf(stderr, "  FYI: -%c=P option not implemented yet!\n", c);
			break;

		case 'C':					// CHECK/STATUS request
		case 'S':					// hidden alias for -C! (-STATUS)
			ctl_type = MPIRWPERF_CHK;
			break;

		case 'R':					// RESET! request
			ctl_type = MPIRWPERF_RESET;
			break;

		case 'v':					// toggle verbose
			verbose = !verbose;
			break;

		case 'h':					// Help! request
			return usage(myname, -1);

		case ':':		/* option without argument */
//			fprintf(stderr, "  Error: %c option requires an argument\n", optopt);
			return usage(myname, optopt-'a'+1);
			break;

		case '?':
//			fprintf(stderr, "  Error: %c option not recognized!\n", optopt);
			return usage(myname, c-'a'+1);
	}

	/* final sanity filter... */
	if ( ((pi->iters/2048) /*=48d=*/ * pi->nblks) > 8000 )
	{
//		fprintf(stderr, "\n        YIKES!  Reset iters from %d to %d!\n",
//				pi->iters, 8000*2048/pi->nblks - (pi->skip*80));
//		fprintf(stderr,   "                (to avoid capacity overflow)\n\n");
		pi->iters = 8000 * 2048 / pi->nblks - (pi->skip*80);
	}

#if 0
	for ( ; optind < argc; optind++) {
		if (access(argv[optind], 4)) {
			...
		}
	}
#endif

	return 0;
}


/*==========================================================================*/

